#include <cstdio>
#include <cstring>
#include <cfloat>
#include <functional>
#include <algorithm>


#include "glutwm.h"
#include "modelwindow.h"
#include "error.h"
#include "cloud.h"
#include "TriMesh.h"
#include "trimeshutil.h"
#include "pointsample.h"
#include "stratify.h"
#include "plysoup.h"
#include "la.h"
#include "vector.h"

TriMesh *mesh;
Cloud cloud;
Array<char> mask;

class BPA {
public:
    struct Edge {
        int i, j, o;
        Vector3 c;
        Edge *next, *previous;
    };
    struct Cell {
        int i;
        Cell *next;
    };
    struct Point {
        char used, front;
    };

    BPA(void) {
        cloud = NULL;
        radius = 1;
    }

    void SetCloud(Cloud *cloud, double radius) {
        this->cloud = cloud;
        this->radius = radius;
        this->vsize = 2*radius;
        BBox3 box;
        for (int i = 0; i < cloud->vertex.GetSize(); i++)
            box.Merge(cloud->vertex[i]);
        Vector3 size = box.GetSize();
        width = (int) (size[0]/vsize) + 3;
        height = (int) (size[1]/vsize) + 3;
        depth = (int) (size[2]/vsize) + 3;
        origin = box.GetMin();
        //debugf(("%dx%dx%d -> %d", width, height, depth, width*height*depth));
        voxel.Resize(width*height*depth);
        for (int i = 0; i < voxel.GetSize(); i++)
            voxel[i] = NULL;
        point.Resize(cloud->vertex.GetSize());
        for (int i = 0; i < cloud->vertex.GetSize(); i++) {
            int v = VoxelIndex(cloud->vertex[i]);
            Cell *cell = (Cell *) malloc(sizeof(Cell));
            cell->i = i;
            cell->next = voxel[v];
            voxel[v] = cell;
            point[i].used = 0;
            point[i].front = 0;
        }
    }

    int VoxelIndex(int x, int y, int z) {
        return z*width*height + y*height + x;
    }

    int VoxelIndex(Vector3 point) {
        int x, y, z;
        VoxelIndex(point, &x, &y, &z);
        return VoxelIndex(x, y, z);
    }

    void VoxelIndex(Vector3 point, int *x, int *y, int *z) {
        point -= origin;
        *x = (int) (point[0]/vsize) + 1;
        *y = (int) (point[1]/vsize) + 1;
        *z = (int) (point[2]/vsize) + 1;
    }

    int Used(int k) {
        return point[k].used;
    }

    int Front(int k) {
        return point[k].front;
    }

    Edge *PopEdge(Edge *&F) {
        if (!F) return NULL;
        Edge *E = F;
        F = F->next;
        E->next = NULL;
        //debugf(("pop: (%d, %d) -> %p", E->i, E->j, E));
        return E;
    }

    void DoneWithEdge(Edge *E) {
        //debugf(("kill: (%d, %d) -> %p", E->i, E->j, E));
        point[E->i].front--;
        point[E->j].front--;
        free(E);
    }

    void MarkAsBoundaryEdge(Edge *E) {
        //debugf(("boundary: (%d, %d) -> %p", E->i, E->j, E));
        point[E->i].front--;
        point[E->j].front--;
        free(E);
    }

    Edge *PushEdge(int i, int j, int o, Vector3 c, Edge *&F) {
        Edge *E = (Edge *) malloc(sizeof(Edge));
        //debugf(("new: (%d, %d) -> %p", i, j, E));
        E->i = i; 
        E->j = j;
        E->o = o;
        E->c = c;
        E->previous = NULL;
        E->next = F;
        if (F) F->previous = E;
        F = E;
        point[i].front++;
        point[j].front++;
        return E;
    }

    void KickEdge(Edge *E, Edge *&F) {
        if (E->previous != NULL) E->previous->next = E->next;
        if (E->next != NULL) E->next->previous = E->previous;
        if (F == E) F = E->next;
        point[E->i].front--;
        point[E->j].front--;
        //debugf(("kick: (%d, %d) -> %p", E->i, E->j, E));
        E->next = NULL;
        free(E);
    }

    Edge *FindEdge(int i, int j, Edge *&F) {
        Edge *E = F;
        while (E) {
            if (E->i == i && E->j == j) break;
            E = E->next;
        }
        //debugf(("finding: (%d, %d) -> %p", i, j, E));
        return E;
    }

    void Triangulate(Soup *soup) {
        soup->vertex = cloud->vertex;
        soup->normal = cloud->normal;
        Edge *F = NULL;
        int count = 0;
        while (1) {
            while (1) {
                Edge *Eij = PopEdge(F);
                if (!Eij) break;
                count -= 1;
                Vector3 c;
                debugf(("before pivot %d", count));
                int k = BallPivot(Eij, &c);
                if (k >= 0 && (!Used(k) || Front(k))) {
                    OutputTriangle(Eij->i, k, Eij->j, soup);
                    Edge *Eik = PushEdge(Eij->i, k, Eij->j, c, F);
                    Edge *Ekj = PushEdge(k, Eij->j, Eij->i, c, F);
                    Edge *Eki = FindEdge(k, Eij->i, F);
                    Edge *Ejk = FindEdge(Eij->j, k, F);
                    count += 2;
                    if (Eki) {
                        KickEdge(Eki, F);
                        KickEdge(Eik, F);
                        count -= 2;
                    }
                    if (Ejk) {
                        KickEdge(Ejk, F);
                        KickEdge(Ekj, F);
                        count -= 2;
                    }
                    DoneWithEdge(Eij);
                } else {
                    //debugf(("%d: (%d, %d) not good", k, Used(k), Front(k)));
                    MarkAsBoundaryEdge(Eij);
                }
            }
            int i, j, k;
            Vector3 c;
            debugf(("before find seed %d", count));
            if (FindSeedTriangle(&i, &j, &k, &c)) {
                OutputTriangle(i, j, k, soup);
                PushEdge(i, j, k, c, F);
                PushEdge(j, k, i, c, F);
                PushEdge(k, i, j, c, F);
                count += 3;
            } else break;
        }
    }

    struct nearer : public binary_function<int, int, bool> {
        public:
            nearer(Cloud *cloud, Vector3 point) {
                this->cloud = cloud;
                this->point = point;
            }
            bool operator()(int i, int j) { 
                return Norm2(cloud->vertex[i] - point) < 
                    Norm2(cloud->vertex[j] - point);
            }
        private:
            Cloud *cloud;
            Vector3 point;
    };

    void GetNeighbors(Vector3 pi, Array<int> &neighbor) {
        neighbor.Resize(0);
        int x, y, z;
        VoxelIndex(pi, &x, &y, &z); 
        // collect all neighbors
        for (int m = -1; m <= 1; m++) {
            for (int n = -1; n <= 1; n++) {
                for (int p = -1; p <= 1; p++) {
                    if (x+m <= 0 || x+m >= width-1) continue;
                    if (y+n <= 0 || y+n >= height-1) continue;
                    if (z+p <= 0 || z+p >= depth-1) continue;
                    Cell *cell = voxel[VoxelIndex(x+m, y+n, z+p)];
                    while (cell) {
                        Vector3 pc = cloud->vertex[cell->i];
                        if (Norm2(pc - pi) < vsize*vsize)
                            neighbor.PushBack(cell->i);
                        cell = cell->next;
                    }
                }
            }
        }
    }

    bool FindSeedTriangle(int *r, int *s, int *t, Vector3 *o) {
        Array<int> neighbor;
        for (int i = 0; i < point.GetSize(); i++) {
            if (!point[i].used) {
                Vector3 pi = cloud->vertex[i];
                GetNeighbors(pi, neighbor);
                debugf(("neighbors %d", neighbor.GetSize()));
                // sort acoording to distance
                sort(neighbor.begin(), neighbor.end(), nearer(cloud, pi));
                // check all triangles for a good one
                for (int nj = 0; nj < neighbor.GetSize(); nj++) {
                    int j = neighbor[nj];
                    if (j == i) continue;
                    for (int nk = 0; nk < neighbor.GetSize(); nk++) {
                        int k = neighbor[nk];
                        if (k == i) continue;
                        if (!point[j].used && !point[k].used && j != k) {
                            Vector3 pj = cloud->vertex[j];
                            Vector3 pk = cloud->vertex[k];
                            // check normals
                            if (!ConsistentNormal(i, j, k)) continue;
                            Vector3 c;
                            // see if sphere that touches them contains no
                            // other points
                            if (BallCenter(pi, pj, pk, radius, &c) &&
                                    EmptyBall(c, radius, i, j, k, neighbor)) {
                                *r = i; *s = j; *t = k; *o = c;
                                return true;
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    bool ConsistentNormal(int i, int j, int k) {
        Vector3 pi = cloud->vertex[i];
        Vector3 pj = cloud->vertex[j];
        Vector3 pk = cloud->vertex[k];
        Vector3 n = Cross(pj - pi, pk - pi);
        return Dot(n, cloud->normal[i]) > 0
            && Dot(n, cloud->normal[j]) > 0
            && Dot(n, cloud->normal[k]) > 0;
    }

    bool EmptyBall(Vector3 c, double r, int i, int j, int k,
            Array<int> &neighbor) {
        double r2 = r*r;
        for (int n = 0; n < neighbor.GetSize(); n++) {
            int p = neighbor[n];
            if (p != i && p != j && p != k)
                if (Norm2(cloud->vertex[p] - c) < r2) return false;
        }
        return true;
    }

    Vector3 CircumCenter(Vector3 p1, Vector3 p2, Vector3 p3) {
        Vector3 p12 = p2 - p1;
        Vector3 p23 = p3 - p2;
        Vector3 p31 = p1 - p3;
        Vector3 m12 = (p2 + p1)/2.0;
        Vector3 m23 = (p3 + p2)/2.0;
        Vector3 m31 = (p1 + p3)/2.0;
        double n12 = Norm2(p12);
        double n23 = Norm2(p23);
        double n31 = Norm2(p31);
        Vector3 h12 = p23 - Dot(p23, p12) * p12 / n12;
        Vector3 h23 = p31 - Dot(p31, p23) * p23 / n23;
        Vector3 h31 = p12 - Dot(p12, p31) * p31 / n31;
        Matrix<9, 6> M;
        Matrix<9, 1> b;
        b.Load(0, 0, -m12); b.Load(3, 0, -m23); b.Load(6, 0, -m31);
        M.Load(0, 3, h12); M.Load(3, 4, h23); M.Load(6, 5, h31);
        for (int i = 0; i < 3; i++)
            for (int j = 0; j < 3; j++) 
                M(i*3+j, j) = -1;
        gels('N', M, b);
        Vector3 c;
        b.Store(0, 0, c); 
        return c;
    }

    bool BallCenter(Vector3 p1, Vector3 p2, Vector3 p3, double r, 
            Vector3 *c) {
        Vector3 cc = CircumCenter(p1, p2, p3);
        double d2 = Norm2(p1 - cc);
        double r2 = r*r;
        Vector3 n = Cross(p2 - p1, p3 - p1);
        n = Normalize(n);
        if (d2 < r2) {
            *c = cc + n*sqrt(r2 - d2);
            return true;
        } else return false;
    }

    void Print(Edge *E) {
        while (E) {
            //debugf(("(%d, %d) -> %p", E->i, E->j, E));
            E = E->next;
        }
    }

    int BallPivot(Edge *Eij, Vector3 *c) {
        Vector3 pi = cloud->vertex[Eij->i];
        Vector3 pj = cloud->vertex[Eij->j];
        Vector3 m = (pi + pj)/2.0;
        Vector3 dr = Eij->c - m;
        Vector3 dij = pj - pi;
        Array<int> neighbor;
        GetNeighbors(m, neighbor); 
        int best_k = -1;
        double best_dot = -DBL_MAX;
        for (int nk = 0; nk < neighbor.GetSize(); nk++) {
            int k = neighbor[nk];
            if (k == Eij->i || k == Eij->j) continue;
            Vector3 pk = cloud->vertex[k];
            Vector3 nc;
            if (BallCenter(pi, pk, pj, radius, &nc)) {
                Vector3 ndr = nc - m;
                double d = Dot(ndr, dr);
                if (d > best_dot && !Negative(Dot(Cross(dr, ndr), dij))) {
                    best_k = k;
                    best_dot = d;
                    *c = nc;
                    //debugf(("%d: (%d %d) accepted", k, Used(k), Front(k)));
                } else {
#if 0
                    debugf(("%d: (%d %d) rejected", k, Used(k), Front(k)));
                    Vector3 cc = Eij->c;
                    debugf(("c: (%g %g %g)", cc[0], cc[1], cc[2]));
                    debugf(("m: (%g %g %g)", m[0], m[1], m[2]));
                    debugf(("dr: (%g %g %g)", dr[0], dr[1], dr[2]));
                    debugf(("ndr: (%g %g %g)", ndr[0], ndr[1], ndr[2]));
                    debugf(("dij: (%g %g %g)", dij[0], dij[1], dij[2]));
                    debugf(("dot-cross: %g", Dot(Cross(dr, ndr), dij)));
                    debugf(("%g vs. best %g", d, best_dot));
#endif
                }
            }
        }
        // debugf(("%d: (%d %d) returned", best_k, Used(best_k), Front(best_k)));
        return best_k;
    }
    void OutputTriangle(int i, int j, int k, Soup *soup) {
//        debugf(("%d -> %d %d %d", soup->triangle.GetSize(), i, j, k));
        soup->triangle.PushBack(Soup::Triangle(i, j, k));
        point[i].used = point[j].used = point[k].used = 1;
    }

    Array<Cell *> voxel;
    int width, height, depth;
    Vector3 origin;
    Array<Point> point;
    Cloud *cloud;
    double radius, vsize;
};

class DebugWindow: public ModelWindow {
public:
    typedef ModelWindow Superclass;
    DebugWindow(void) {
        selected = -1;
        cloud = NULL;
        radius = 1;
    }
    void SetCloud(Cloud *cloud, double radius) {
        this->cloud = cloud;
        this->radius = radius;
        bpa.SetCloud(cloud, radius);
    }
    void Fit(void) {
        BBox3 box;
        for (int i = 0; i < cloud->vertex.GetSize(); i++)
            box.Merge(cloud->vertex[i]);
        SetCenter(box.GetCenter());
        SetSize(Norm(box.GetSize()));
    }
    void DrawModel(void) {
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LESS);
        glClearColor(0, 0, 0, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glPointSize(3);
        glDisable(GL_LIGHTING);
        glBegin(GL_POINTS);
        if (selected >= 0) {
            glColor3d(1, 0, 0);
            glVertex3dv(&cloud->vertex[selected][0]);
            glColor3d(0, 0, 1);
            Array<int> neighbor;
            bpa.GetNeighbors(cloud->vertex[selected], neighbor);
            for (int i = 0; i < neighbor.GetSize(); i++)
                glVertex3dv(&cloud->vertex[neighbor[i]][0]);
        }
        glColor3d(1, 1, 1);
        for (int i = 0; i < cloud->vertex.GetSize(); i++)
            glVertex3dv(&cloud->vertex[i][0]);
        glEnd();
        int i, j, k;
        Vector3 c;
        bpa.FindSeedTriangle(&i, &j, &k, &c);
        Vector3 pi = cloud->vertex[i];
        Vector3 pj = cloud->vertex[j];
        Vector3 pk = cloud->vertex[k];
        glColor3d(0, 1, 0);
        glBegin(GL_TRIANGLES);
            glVertex3dv(&pi[0]);
            glVertex3dv(&pj[0]);
            glVertex3dv(&pk[0]);
        glEnd();
        glColor3d(1, 0, 1);
        glBegin(GL_POINTS);
            glVertex3dv(&c[0]);
        glEnd();
        Vector3 n = Cross(pj - pi, pk - pi);
        n = Normalize(n);
        glBegin(GL_LINES);
            glVertex3dv(&pi[0]);
            glVertex3dv(&(pi + n*radius)[0]);
            glVertex3dv(&pj[0]);
            glVertex3dv(&(pj + n*radius)[0]);
            glVertex3dv(&pk[0]);
            glVertex3dv(&(pk + n*radius)[0]);
        glEnd();
        last = i;
    }
    int PickByColor(int x, int y);
    void ColorSamples(void);
    void Button(int button, int state, int x, int y) {
        if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
            if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
                selected = PickByColor(x, y);
                debugf(("selected: %d", selected));
                PostRedisplay();
            }
        } else Superclass::Button(button, state, x, y);
    }
    void Keyboard(unsigned char c, int x, int y) {
        if (c == 'n') {
            bpa.point[last].used = 1;
            PostRedisplay();
        } else Superclass::Keyboard(c, x, y);
    }
private:
    Cloud *cloud;
    double radius;
    int selected;
    BPA bpa;
    int last;
};

static void i2color(int i, GLubyte *color)
{
    color[0] = (GLubyte) (i & 0xff);
    i >>= 8;
    color[1] = (GLubyte) (i & 0xff);
    i >>= 8;
    color[2] = (GLubyte) (i & 0xff);
}

void
DebugWindow::
ColorSamples(void) {
    glClearColor(1, 1, 1, 1);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glDisable(GL_LIGHTING);
    glDisable(GL_BLEND);
    glPointSize(5);
    glBegin(GL_POINTS);
    // loop through each vertex drawing it in a unique color
    for (size_t i = 0; i < cloud->vertex.size(); i++) {
        GLubyte color[3];
        i2color(i, color);
        glColor3ubv(color);
        glVertex3dv(&cloud->vertex[i][0]);
    }
    glEnd();
}

static int color2i(GLubyte *color)
{
    int i = color[2];
    i <<= 8;
    i |= color[1];
    i <<= 8;
    i |= color[0];
    return (i == 0xffffff) ? -1: i;
}

int
DebugWindow::
PickByColor(int mx, int my)
{
    GLubyte color[3];
    ColorSamples();
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    GlutWMReadPixels(mx, my, 1, 1, GL_RGB, color);
    return color2i(color);
}

PlySoup soup;
DebugWindow wdebug;

int main(int argc, char *argv[])
{
#if 1
    //GlutWMOpen(&argc, argv);
    if (argc < 5) {
        fprintf(stderr, "Usage\n"
			  "pivot <level> <lambda> <distance> <ply> <radius>\n");
        return 1;
    }
    double radius = (argc > 5) ? atof(argv[5]): 1;
    debugf(("Loading mesh"));
    mesh = TriMesh::read(argv[4]);
    if (!mesh) {
        errorf(("Unable to load mesh '%s'", argv[4]));
        return 1;
    }
    debugf(("%d vertices, %d faces", mesh->vertices.size(), 
                mesh->faces.size()));
    debugf(("Generating stratified distribution"));
	int level = atoi(argv[1]);
	double lambda = strtod(argv[2], NULL);
	double distance = strtod(argv[3], NULL);
    StratifiedSample(mesh, level, lambda, distance, &cloud, mask);
    BPA bpa;
    //GlutWMCreateWindow("Debug", &wdebug);
    //wdebug.SetCloud(&cloud, mesh->bsphere.r / pow(2.0, level-3));
    //GlutWMMainLoop();
    bpa.SetCloud(&cloud, radius*mesh->bsphere.r / pow(2.0, level-2));
    bpa.Triangulate(&soup);
    soup.Store("ball.ply");
#else
    for (int i = 0; i < 3; i++) {
        for (int j = 0; j < 3; j++) {
            cloud.vertex.PushBack(Vector3(i, j, 0));
            cloud.normal.PushBack(Vector3(0, 0, 1));
        }
    }
    BPA bpa;
    bpa.SetCloud(&cloud, argc > 1 ? atof(argv[1]): 1);
    bpa.Triangulate(&soup);
    soup.Store("ball.ply");
#endif
    return 0;
}
